@@ -1,4 +1,4 @@
-# Copyright (C) 2003-2016 GraphicsMagick Group
+# Copyright (C) 2003-2023 GraphicsMagick Group
 # Copyright (C) 2002 ImageMagick Studio
 # Copyright (C) 1999 E. I. du Pont de Nemours and Company
 #
@@ -192,16 +192,17 @@ sub testCompositeCompare {
 }
 
 #
-# Test reading a file in which several signatures are possible,
-# depending on available image depth.
+# Test reading a file (and blob) in which several signatures are
+# possible, depending on available image depth.
 #
-# Usage: testRead( read filename, expected ref_8 [, expected ref_16] [, expected ref_32] );
+# Usage: testRead( read filename, read options, expected ref_8 [, expected ref_16] [, expected ref_32] );
 #
 sub testRead {
-  my( $infile, $ref_8, $ref_16, $ref_32 ) =  @_;
+  my( $infile, $read_options, $ref_8, $ref_16, $ref_32 ) =  @_;
 
-  my($image,$magick,$success,$depth,$ref_signature);
+  my($image,$magick,$success,$depth,$ref_signature,$errorinfo,$failure,$status);
 
+  $errorinfo='';
   $failure=0;
 
   if ( !defined( $ref_16 ) )
@@ -237,7 +238,18 @@ sub testRead {
 
     print( "  testing reading from file \"", $infile, "\" ...\n");
     $image=Graphics::Magick->new;
-    $image->Set(size=>'512x512');
+    #$image->Set(size=>'512x512');
+
+    if ( "$read_options" ne "" ) {
+        eval "\$status=\$image->Set($read_options);";
+        if ("$status")
+        {
+            $errorinfo = "Set($read_options): $status";
+            warn("$errorinfo");
+            goto READ_RUNTIME_ERROR;
+        }
+    }
+
     $status=$image->ReadImage("$infile");
     if( "$status" && !($status =~ /Exception ((315)|(350))/)) {
       print "ReadImage $infile: $status\n";
@@ -249,12 +261,12 @@ sub testRead {
       undef $status;
       $magick=$image->Get('magick');
       $signature=$image->Get('signature');
-      
+
       if ( $signature ne $ref_signature ) {
         print "ReadImage()\n";
-	print "Image: $infile, signatures do not match.\n";
-	print "     Computed: $signature\n";
-	print "     Expected: $ref_signature\n";
+        print "Image: $infile, signatures do not match.\n";
+        print "     Computed: $signature\n";
+        print "     Expected: $ref_signature\n";
         print "     Depth:    $depth\n";
         ++$failure;
         defined $ENV{'PERL_DEBUG'} && $image->Display();
@@ -278,6 +290,17 @@ sub testRead {
         close( FILE );
         if( defined( $blob ) ) {
           $image=Graphics::Magick->new(magick=>$magick);
+          #$image->Set(size=>'512x512');
+
+          if ( defined( $read_options ) && "$read_options" ne "" ) {
+              eval "\$status=\$image->Set($read_options);";
+              if ("$status")
+              {
+                  $errorinfo = "Set($read_options): $status";
+                  warn("$errorinfo");
+                  goto READ_RUNTIME_ERROR;
+              }
+          }
           $status=$image->BlobToImage( $blob );
           undef $blob;
           if( "$status" && !($status =~ /Exception ((315)|(350))/)) {
@@ -303,6 +326,8 @@ sub testRead {
     undef $image;
   }
 
+ READ_RUNTIME_ERROR:
+
   #
   # Display test status
   #
@@ -310,12 +335,13 @@ sub testRead {
     print "not ok $test\n";
   } else {
     print "ok $test\n";
-  }    
+  }
 }
 
 
 #
 # Test reading a file, and compare with a reference file
+# This one does not test reading from an in-memory blob
 #
 sub testReadCompare {
   my( $srcimage_name,$refimage_name, $read_options,
@@ -326,7 +352,7 @@ sub testReadCompare {
 
   # Create images
   $srcimage=Graphics::Magick->new;
-  $refimage=Graphics::Magick->new;  
+  $refimage=Graphics::Magick->new;
 
   if ( "$read_options" ne "" ) {
     eval "\$status=\$srcimage->Set($read_options);";
@@ -337,7 +363,7 @@ sub testReadCompare {
         goto COMPARE_RUNTIME_ERROR;
       }
   }
-  
+
   $status=$srcimage->ReadImage("$srcimage_name");
   if ("$status")
     {
@@ -436,11 +462,11 @@ sub testReadCompare {
 # Test reading a file which requires a file size to read (GRAY, RGB, CMYK)
 # or supports multiple resolutions (JBIG, JPEG, PCD)
 #
-# Usage: testRead( read filename, size, depth, expected ref_8 [, expected ref_16] );
+# Usage: testReadSized( read filename, size, depth, expected ref_8 [, expected ref_16] );
 #
 sub testReadSized {
   my( $infile, $size, $depth, $ref_8, $ref_16, $ref_32 ) =  @_;
-  
+
   my($image,$ref_signature);
 
   if ( !defined( $ref_16 ) )
@@ -486,9 +512,9 @@ sub testReadSized {
     $signature=$image->Get('signature');
       if ( $signature ne $ref_signature ) {
         print "ReadImage()\n";
-	print "Image: $infile, signatures do not match.\n";
-	print "     Computed: $signature\n";
-	print "     Expected: $ref_signature\n";
+        print "Image: $infile, signatures do not match.\n";
+        print "     Computed: $signature\n";
+        print "     Expected: $ref_signature\n";
         print "     Depth:    $depth\n";
         print "not ok $test\n";
         defined $ENV{'PERL_DEBUG'} && $image->Display();
@@ -515,7 +541,7 @@ sub testReadSized {
 #
 sub testReadWrite {
   my( $infile, $outfile, $writeoptions, $ref_8, $ref_16, $ref_32 ) = @_;
-  
+
   my($image);
 
   if ( !defined( $ref_16 ) )
@@ -603,7 +629,7 @@ sub testReadWriteCompare {
   $errorinfo='';
 
   $image=Graphics::Magick->new;
-  $refimage=Graphics::Magick->new;  
+  $refimage=Graphics::Magick->new;
 
   #
   # Read the initial image
@@ -735,7 +761,7 @@ sub testReadWriteCompare {
 #
 # Test writing a file by first reading a source image, writing to a
 # new image, and reading the written image.  Depends on detecting
-# reported errors by ImageMagick
+# reported errors by GraphicsMagick
 #
 # Usage: testReadWrite( read filename, write filename, write options);
 #
@@ -749,9 +775,9 @@ sub testReadWriteCompare {
 #
 sub testReadWriteNoVerify {
   my( $infile, $outfile, $writeoptions) = @_;
-  
+
   my($image, $images);
-  
+
   $image=Graphics::Magick->new;
   $status=$image->ReadImage("$infile");
   if( "$status" ) {
@@ -810,7 +836,7 @@ sub testReadWriteNoVerify {
 sub testReadWriteSized {
   my( $infile, $outfile, $size, $readdepth, $writeoptions, $ref_8, $ref_16,
       $ref_32 ) = @_;
-  
+
   my($image,$depth,$ref_signature);
 
   if ( !defined( $ref_16 ) )
@@ -917,7 +943,7 @@ sub testSetAttribute {
   my( $srcimage, $name, $attribute ) = @_;
 
   my($image);
-  
+
   # Create temporary image
   $image=Graphics::Magick->new;
 
@@ -1031,7 +1057,7 @@ sub testMontage {
                  '#808080', '#808000', '#FFFFFF', '#FFFF00',
                  '#800000', '#000080', '#FF0000', '#0000FF',
                  '#800080', '#008080', '#FF00FF', '#00FFFF' );
-  
+
   my $color;
   foreach $color ( @colors ) {
 
@@ -1068,7 +1094,7 @@ sub testMontage {
     print "not ok $test\n";
     return 1;
   }
-  
+
   if( ! ref($montage) ) {
     print "not ok $test\n";
   } else {
@@ -1079,12 +1105,12 @@ sub testMontage {
       if ( $signature ne $ref_signature ) {
         print "ReadImage()\n";
         print "Test $test, signatures do not match.\n";
-	print "     Computed: $signature\n";
-	print "     Expected: $ref_signature\n";
+        print "     Computed: $signature\n";
+        print "     Expected: $ref_signature\n";
         print "     Depth:    $depth\n";
         $status = $montage->Write("test_${test}_out.miff");
         warn "Write: $status" if "$status";
-          
+
         print "not ok $test\n";
       } else {
         # Check montage directory
