@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003-2022 GraphicsMagick Group
+% Copyright (C) 2003-2023 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 %
 % This program is covered by multiple licenses, which are described in
@@ -21,7 +21,7 @@
 %                                                                             %
 %                              Software Design                                %
 %                              Jaroslav Fojtik                                %
-%                              June 2000 - 2022                               %
+%                              June 2000 - 2023                               %
 %                         Rework for GraphicsMagick                           %
 %                              Bob Friesenhahn                                %
 %                               Feb-May 2003                                  %
@@ -46,6 +46,7 @@
 #include "magick/shear.h"
 #include "magick/tempfile.h"
 #include "magick/transform.h"
+#include "magick/quantize.h"
 #include "magick/utility.h"
 
 
@@ -189,9 +190,19 @@ static const RGB_Record WPG1_Palette[256]={
 };
 
 
-static int ApproveFormatForWPG(const char *Format)
+static int ApproveFormatForWPG(const char *Format, ExceptionInfo *exception)
 {
+  const MagickInfo *magick_info;
+
   if(!strcmp(Format,"PFB")) return 0;
+
+  /*if(!strcmp(Format,"8BIMTEXT")) return 0; This test is no longer needed, META module includes this case. */
+  magick_info = GetMagickInfo(Format,exception);
+  if(magick_info != (const MagickInfo *)NULL)
+  {
+    if(strcmp(magick_info->module, "META") == 0) return 0;
+  }
+
   return 1;
 }
 
@@ -235,7 +246,7 @@ static unsigned int IsWPG(const unsigned char *magick,const size_t length)
   return(False);
 }
 
-static int Rd_WP_DWORD(Image *image, unsigned long *d)
+static int Rd_WP_DWORD(Image *image, magick_uint32_t *d)
 {
   unsigned char b;
 
@@ -419,7 +430,7 @@ static MagickPassFail ZeroFillMissingData(unsigned char *BImgBuff,unsigned long
 
 
 /* WPG1 raster reader.
- * @return      0 - OK; -2 - alocation failure; -3 unaligned column; -4 - image row overflowl
+ * @return      0 - OK; -2 - allocation failure; -3 unaligned column; -4 - image row overflowl
                 -5 - blob read error; -6 - row insert problem  */
 static int UnpackWPGRaster(Image *image,int bpp)
 {
@@ -475,7 +486,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
           }
         }
       else {
-        if(RunCount)   /* next runcount byte are readed directly */
+        if(RunCount)   /* next runcount byte are readd directly */
           {
             for(i=0;i < (int) RunCount;i++)
               {
@@ -864,7 +875,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   /*
     Verify if this is an allowed subordinate image format
   */
-  if(!ApproveFormatForWPG(format))
+  if(!ApproveFormatForWPG(format,exception))
   {
     (void) LogMagickEvent(CoderEvent, GetMagickModule(),
                         "Format \"%s\" cannot be embedded inside WPG.", format);
@@ -962,8 +973,18 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
     {
       goto FINISH_UNL;
     }
-  if(exception->severity>=ErrorException) /* When exception is raised, destroy image2 read. */
+  if(exception->severity >= ErrorException) /* When exception is raised, destroy image2 read. */
   {
+    if(image->logging)
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(), "Exception raised during embedded image reading.");
+    CloseBlob(image2);
+    DestroyImageList(image2);
+    goto FINISH_UNL;
+  }
+  if(!GetPixelCachePresent(image2))
+  {
+    if(image->logging)
+      (void) LogMagickEvent(CoderEvent,GetMagickModule(), "Pixel cache is missing for embedded image.");
     CloseBlob(image2);
     DestroyImageList(image2);
     goto FINISH_UNL;
@@ -1015,6 +1036,21 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   return(image);
 }
 
+static int EnsureNextImage(const ImageInfo *image_info, Image **pp_image)
+{
+  if(pp_image==NULL) return -1;
+  if(*pp_image==NULL) return -2;
+  if(image_info==NULL) return -3;
+
+  AllocateNextImage(image_info,*pp_image);
+  if((*pp_image)->next == (Image *) NULL) return -4;
+  *pp_image = SyncNextImageInList(*pp_image);
+  (*pp_image)->columns = (*pp_image)->rows = 0;
+  (*pp_image)->colors = 0;
+return 0;
+}
+
+
 #define LogHeaderWPG(_WPG_HEADER) \
   (void)LogMagickEvent(CoderEvent,GetMagickModule(), \
                        "WPG Header Id=%Xh:\n" \
@@ -1037,12 +1073,12 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   (void)LogMagickEvent(CoderEvent,GetMagickModule(), \
                        "WPG Bitmap1 Header:\n" \
                        "    Width=%u\n" \
-                       "    Heigth=%u\n" \
+                       "    Height=%u\n" \
                        "    Depth=%u\n" \
                        "    HorzRes=%u\n" \
                        "    VertRes=%u", \
                             (unsigned int)_WPG_BITMAP_TYPE1.Width, \
-                            (unsigned int)_WPG_BITMAP_TYPE1.Heigth, \
+                            (unsigned int)_WPG_BITMAP_TYPE1.Height, \
                             (unsigned int)_WPG_BITMAP_TYPE1.Depth, \
                             (unsigned int)_WPG_BITMAP_TYPE1.HorzRes, \
                             (unsigned int)_WPG_BITMAP_TYPE1.VertRes)
@@ -1056,7 +1092,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
                        "    UpRightX=%u\n" \
                        "    UpRightY=%u\n" \
                        "    Width=%u\n" \
-                       "    Heigth=%u\n" \
+                       "    Height=%u\n" \
                        "    Depth=%u\n" \
                        "    HorzRes=%u\n" \
                        "    VertRes=%u", \
@@ -1066,12 +1102,32 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
                             (unsigned int)_WPG_BITMAP_TYPE2.UpRightX, \
                             (unsigned int)_WPG_BITMAP_TYPE2.UpRightY, \
                             (unsigned int)_WPG_BITMAP_TYPE2.Width, \
-                            (unsigned int)_WPG_BITMAP_TYPE2.Heigth, \
+                            (unsigned int)_WPG_BITMAP_TYPE2.Height, \
                             (unsigned int)_WPG_BITMAP_TYPE2.Depth, \
                             (unsigned int)_WPG_BITMAP_TYPE2.HorzRes, \
                             (unsigned int)_WPG_BITMAP_TYPE2.VertRes)
 
 
+typedef struct
+{
+    magick_uint16_t StartIndex;
+    magick_uint16_t NumOfEntries;
+} WPGColorMapRec;
+
+
+static void LoadPaletteRec(Image *image, WPGColorMapRec *pWPG_Palette, const int logging)
+{
+  pWPG_Palette->StartIndex=ReadBlobLSBShort(image);
+  pWPG_Palette->NumOfEntries=ReadBlobLSBShort(image);
+  if(logging)
+    (void)LogMagickEvent(CoderEvent,GetMagickModule(),
+                "WPG Color palette:\n"
+                "    StartIndex=%u\n"
+                "    NumOfEntries=%u\n", (unsigned)pWPG_Palette->StartIndex, (unsigned)pWPG_Palette->NumOfEntries);
+
+}
+
+
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %                                                                             %
@@ -1109,8 +1165,8 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   {
     unsigned long FileId;
     ExtendedSignedIntegralType DataOffset;  /* magick_uint32_t */
-    unsigned int ProductType;
-    unsigned int FileType;
+    unsigned char ProductType;
+    unsigned char FileType;
     unsigned char MajorVersion;
     unsigned char MinorVersion;
     unsigned int EncryptKey;
@@ -1120,15 +1176,15 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   typedef struct
   {
     unsigned char RecType;
-    unsigned long RecordLength;
+    magick_uint32_t RecordLength;
   } WPGRecord;
 
   typedef struct
   {
     unsigned char Class;
     unsigned char RecType;
-    unsigned long Extension;
-    unsigned long RecordLength;
+    magick_uint32_t Extension;
+    magick_uint32_t RecordLength;
   } WPG2Record;
 
   typedef struct
@@ -1141,7 +1197,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   typedef struct
   {
     unsigned int Width;
-    unsigned int Heigth;
+    unsigned int Height;
     unsigned int Depth;
     unsigned int HorzRes;
     unsigned int VertRes;
@@ -1150,7 +1206,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   typedef struct
   {
     unsigned int Width;
-    unsigned int Heigth;
+    unsigned int Height;
     unsigned char Depth;
     unsigned char Compression;
   } WPG2BitmapType1;
@@ -1163,18 +1219,12 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
     unsigned int UpRightX;
     unsigned int UpRightY;
     unsigned int Width;
-    unsigned int Heigth;
+    unsigned int Height;
     unsigned int Depth;
     unsigned int HorzRes;
     unsigned int VertRes;
   } WPGBitmapType2;
 
-  typedef struct
-  {
-    unsigned int StartIndex;
-    unsigned int NumOfEntries;
-  } WPGColorMapRec;
-
   /*
   typedef struct {
     unsigned long PS_unknown1;
@@ -1227,6 +1277,10 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   BlobInfo *TmpBlob;
   magick_off_t FilePos, filesize;
 
+  unsigned char *pPalette = NULL;
+  magick_uint16_t PaletteItems = 0;
+  magick_uint32_t PaletteAllocBytes = 0;
+
   tCTM CTM;         /*current transform matrix*/
 
   /*
@@ -1240,6 +1294,8 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   logging = LogMagickEvent(CoderEvent,GetMagickModule(),"enter");
 
   image=AllocateImage(image_info);
+  image->rows=0;
+  image->columns=0;
   image->depth=8;
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == False)
@@ -1250,15 +1306,15 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   */
   Header.FileId=ReadBlobLSBLong(image);
   Header.DataOffset=(ExtendedSignedIntegralType) ReadBlobLSBLong(image);
-  Header.ProductType=ReadBlobLSBShort(image);
-  Header.FileType=ReadBlobLSBShort(image);
+  Header.ProductType=ReadBlobByte(image);
+  Header.FileType=ReadBlobByte(image);
   Header.MajorVersion=ReadBlobByte(image);
   Header.MinorVersion=ReadBlobByte(image);
   Header.EncryptKey=ReadBlobLSBShort(image);
   Header.Reserved=ReadBlobLSBShort(image);
   if(logging) LogHeaderWPG(Header);
 
-  if (Header.FileId!=0x435057FF || (Header.ProductType>>8)!=0x16)
+  if (Header.FileId!=0x435057FF || Header.FileType!=0x16)
     ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
   if (Header.EncryptKey!=0)
     ThrowReaderException(CoderError,EncryptedWPGImageFileNotSupported,image);
@@ -1285,7 +1341,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
     ThrowReaderException(CorruptImageError,AnErrorHasOccurredReadingFromFile,image);
   }
 
-  switch(Header.FileType)
+  switch(Header.MajorVersion)
     {
     case 1:     /* WPG level 1 */
       BitmapHeader2.RotAngle = 0;
@@ -1301,7 +1357,10 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
 
           FilePos += Rd_WP_DWORD(image,&Rec.RecordLength);
           if((magick_off_t)Rec.RecordLength > filesize)
+          {
+            MagickFreeResourceLimitedMemory(pPalette);     
             ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+          }
           if(EOFBlob(image)) break;
 
           FilePos += (magick_off_t)Rec.RecordLength;
@@ -1321,14 +1380,23 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
             {
             case 0x0B: /* bitmap type 1 */
               BitmapHeader1.Width=ReadBlobLSBShort(image);
-              BitmapHeader1.Heigth=ReadBlobLSBShort(image);
-              if ((BitmapHeader1.Width == 0) || (BitmapHeader1.Heigth == 0))
+              BitmapHeader1.Height=ReadBlobLSBShort(image);
+              if ((BitmapHeader1.Width == 0) || (BitmapHeader1.Height == 0))
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
                 ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+              }
               BitmapHeader1.Depth=ReadBlobLSBShort(image);
               BitmapHeader1.HorzRes=ReadBlobLSBShort(image);
               BitmapHeader1.VertRes=ReadBlobLSBShort(image);
               if(logging) LogWPGBitmapType1(BitmapHeader1);
 
+              if(image->rows!=0 && image->columns!=0)
+              {                       /* Allocate next image structure. */
+                if(EnsureNextImage(image_info, &image) < 0)
+                    goto Finish;
+              }
+
               if(BitmapHeader1.HorzRes && BitmapHeader1.VertRes)
                 {
                   image->units=PixelsPerCentimeterResolution;
@@ -1336,28 +1404,50 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   image->y_resolution=BitmapHeader1.VertRes/470.0;
                 }
               image->columns=BitmapHeader1.Width;
-              image->rows=BitmapHeader1.Heigth;
+              image->rows=BitmapHeader1.Height;
               bpp=BitmapHeader1.Depth;
-				// Whole palette is useless for bilevel image.
-              if(bpp==1) image->storage_class=DirectClass;
+              if (image->logging)
+                (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                      "Image dimensions %lux%lu, bpp=%u", image->columns, image->rows, bpp);
+				/* Whole palette is useless for bilevel image. */
+              if(bpp==1)
+              {
+                image->storage_class = PseudoClass;
+                image->colors = 2;
+                if (!AllocateImageColormap(image,2)) goto NoMemory;
+                image->colormap[0].red = image->colormap[0].green = image->colormap[0].blue = 0;
+                image->colormap[1].red = image->colormap[1].green = image->colormap[1].blue = MaxRGB;
+                image->colormap[0].opacity = image->colormap[1].opacity = OpaqueOpacity;
+                goto UnpackRaster1bpp;	/* bypass cached palette for 1bpp. */
+              }
               goto UnpackRaster;
 
             case 0x0E:  /*Color palette */
-              WPG_Palette.StartIndex=ReadBlobLSBShort(image);
-              WPG_Palette.NumOfEntries=ReadBlobLSBShort(image);
-
-              image->colors=WPG_Palette.NumOfEntries;
-              if (!AllocateImageColormap(image,image->colors))
-                goto NoMemory;
-              image->storage_class = PseudoClass;
-              for (i=WPG_Palette.StartIndex;
-                   i < (int)WPG_Palette.NumOfEntries; i++)
+              LoadPaletteRec(image,&WPG_Palette,logging);
+              PaletteItems = WPG_Palette.NumOfEntries;
+              if((magick_uint32_t)PaletteItems + WPG_Palette.StartIndex > 256)
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(CorruptImageError,ColormapExceedsColorsLimit,image);
+              }
+              if(pPalette==NULL)
+              {
+                PaletteAllocBytes = 3*256;
+                pPalette = MagickAllocateResourceLimitedMemory(unsigned char *,(size_t)PaletteAllocBytes);
+                if(pPalette==NULL)
+                    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+                for(i=0; i<=255; i++)
                 {
-                  image->colormap[i].red=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].green=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].blue=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].opacity = OpaqueOpacity;
+                  pPalette[3*i] = WPG1_Palette[i].Red;
+                  pPalette[3*i+1] = WPG1_Palette[i].Green;
+                  pPalette[3*i+2] = WPG1_Palette[i].Blue;
                 }
+              }
+              if(ReadBlob(image,(size_t) PaletteItems*3,pPalette+((size_t)3*WPG_Palette.StartIndex)) != (size_t) PaletteItems*3)
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+              }
               break;
 
             case 0x11:  /* Start PS l1 */
@@ -1374,14 +1464,23 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               BitmapHeader2.UpRightX=ReadBlobLSBShort(image);
               BitmapHeader2.UpRightY=ReadBlobLSBShort(image);
               BitmapHeader2.Width=ReadBlobLSBShort(image);
-              BitmapHeader2.Heigth=ReadBlobLSBShort(image);
-              if ((BitmapHeader2.Width == 0) || (BitmapHeader2.Heigth == 0))
+              BitmapHeader2.Height=ReadBlobLSBShort(image);
+              if ((BitmapHeader2.Width == 0) || (BitmapHeader2.Height == 0))
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
                 ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+              }
               BitmapHeader2.Depth=ReadBlobLSBShort(image);
               BitmapHeader2.HorzRes=ReadBlobLSBShort(image);
               BitmapHeader2.VertRes=ReadBlobLSBShort(image);
               if(logging) LogWPGBitmapType2(BitmapHeader2);
 
+              if(image->rows!=0 && image->columns!=0)
+              {                       /* Allocate next image structure. */
+                if(EnsureNextImage(image_info, &image) < 0)
+                    goto Finish;
+              }
+
               image->units=PixelsPerCentimeterResolution;
               image->page.width=(unsigned int)
                 ((BitmapHeader2.LowLeftX-BitmapHeader2.UpRightX)/470.0);
@@ -1395,12 +1494,39 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   image->y_resolution=BitmapHeader2.VertRes/470.0;
                 }
               image->columns=BitmapHeader2.Width;
-              image->rows=BitmapHeader2.Heigth;
+              image->rows=BitmapHeader2.Height;
               bpp=BitmapHeader2.Depth;
+              if (image->logging)
+                (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                      "Image dimensions %lux%lu, bpp=%u", image->columns, image->rows, bpp);
 
 UnpackRaster:
               if(bpp>24)
-                {ThrowReaderException(CoderError,ColorTypeNotSupported,image)}
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(CoderError,ColorTypeNotSupported,image);
+              }
+
+              if(pPalette!=NULL && PaletteItems>0)
+              {
+                if(bpp>=16 || PaletteItems<(1<<bpp))
+                  image->colors = PaletteItems;	/*WPG_Palette.NumOfEntries;*/
+                else
+                  image->colors = 1 << bpp;
+                if (!AllocateImageColormap(image,image->colors))
+                  goto NoMemory;
+                image->storage_class = PseudoClass;
+                for (i=0; i<(int)image->colors; i++)
+                {
+                  image->colormap[i].red = ScaleCharToQuantum(pPalette[3*i]);
+                  image->colormap[i].green=ScaleCharToQuantum(pPalette[3*i+1]);
+                  image->colormap[i].blue=ScaleCharToQuantum(pPalette[3*i+2]);
+                  image->colormap[i].opacity = OpaqueOpacity;
+                }
+              }
+
+UnpackRaster1bpp:
+              image->depth = bpp;
 
               if ((image->storage_class != PseudoClass) && (bpp != 24) && bpp!=1)
                 {
@@ -1408,8 +1534,8 @@ UnpackRaster:
                   if (!AllocateImageColormap(image,image->colors))
                     {
                     NoMemory:
-                      ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,
-                                           image)
+                      MagickFreeResourceLimitedMemory(pPalette);
+                      ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image)
                     }
                   image->storage_class = PseudoClass;
                   /* printf("Load default colormap \n"); */
@@ -1432,8 +1558,8 @@ UnpackRaster:
               if(bpp == 1)
                 {
                   if(image->colors<=0)
-                                  {
-                                image->colormap[0].red =
+                    {
+                      image->colormap[0].red =
                         image->colormap[0].green =
                         image->colormap[0].blue = 0;
                       image->colormap[0].opacity = OpaqueOpacity;
@@ -1453,6 +1579,7 @@ UnpackRaster:
                 if(UnpackWPGRaster(image,bpp) < 0)
                 {               /* The raster cannot be unpacked */
                 DecompressionFailed:
+                  MagickFreeResourceLimitedMemory(pPalette);
                   ThrowReaderException(CoderError,UnableToDecompressImage,image)
                 }
 
@@ -1505,14 +1632,6 @@ UnpackRaster:
                 if (image->scene >= (image_info->subimage+image_info->subrange-1))
                   goto Finish;
 
-              /* Allocate next image structure. */
-              AllocateNextImage(image_info,image);
-              image->depth=8;
-              if (image->next == (Image *) NULL)
-                goto Finish;
-              image=SyncNextImageInList(image);
-              image->columns=image->rows=0;
-              image->colors=0;
               break;
 
             case 0x1B:  /* Postscript l2 */
@@ -1558,44 +1677,77 @@ UnpackRaster:
               StartWPG.VerticalUnits=ReadBlobLSBShort(image);
               StartWPG.PosSizePrecision=ReadBlobByte(image);
               break;
+
             case 0x0C:    /* Color palette */
-              WPG_Palette.StartIndex=ReadBlobLSBShort(image);
-              WPG_Palette.NumOfEntries=ReadBlobLSBShort(image);
+              LoadPaletteRec(image,&WPG_Palette,logging);
 
               /* Sanity check for amount of palette entries. */
               if (WPG_Palette.NumOfEntries == 0)
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
                 ThrowReaderException(CorruptImageError,UnrecognizedNumberOfColors,image);
-
-              if (WPG_Palette.NumOfEntries > MaxMap+1)
+              }
+              if ((unsigned) (WPG_Palette.NumOfEntries-1) > MaxMap)
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
                 ThrowReaderException(CorruptImageError,ColormapExceedsColorsLimit,image);
+              }
 
               if ( (WPG_Palette.StartIndex > WPG_Palette.NumOfEntries) ||
-                   (((WPG_Palette.NumOfEntries-WPG_Palette.StartIndex) >
-                     ((Rec2.RecordLength-2-2) / 3))) )
-                 ThrowReaderException(CorruptImageError,InvalidColormapIndex,image);
+                   ((((magick_int32_t)WPG_Palette.NumOfEntries-(magick_int32_t)WPG_Palette.StartIndex) >
+                   (((magick_int32_t)Rec2.RecordLength-2-2) / 4))) )
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(CorruptImageError,InvalidColormapIndex,image);
+              }
 
-              image->colors=WPG_Palette.NumOfEntries;
-              if (!AllocateImageColormap(image,image->colors))
-                ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
-
-              for (i=WPG_Palette.StartIndex;
-                   i < (int)WPG_Palette.NumOfEntries; i++)
+              if(pPalette!=NULL &&
+                 PaletteAllocBytes < 4*(WPG_Palette.StartIndex+WPG_Palette.NumOfEntries))
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                PaletteAllocBytes = 0;
+              }
+              if(pPalette==NULL)
+              {
+                PaletteItems = WPG_Palette.NumOfEntries;
+                PaletteAllocBytes = 4*(WPG_Palette.StartIndex+WPG_Palette.NumOfEntries);
+                if(PaletteAllocBytes < 4*256) PaletteAllocBytes = 4*256;
+                pPalette = MagickAllocateResourceLimitedMemory(unsigned char *,(size_t)PaletteAllocBytes);
+                if(pPalette==NULL)
+                    ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+                for(i=0; i<=255; i++)
                 {
-                  image->colormap[i].red=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].green=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].blue=ScaleCharToQuantum(ReadBlobByte(image));
-                  image->colormap[i].opacity = OpaqueOpacity;
-                  (void) ReadBlobByte(image);   /*Opacity??*/
+                  pPalette[4*i] = WPG1_Palette[i].Red;
+                  pPalette[4*i+1] = WPG1_Palette[i].Green;
+                  pPalette[4*i+2] = WPG1_Palette[i].Blue;
+                  pPalette[4*i+3] = OpaqueOpacity;
                 }
+              }
+              if(ReadBlob(image,(size_t) PaletteItems*4,pPalette+((size_t)4*WPG_Palette.StartIndex)) != (size_t) PaletteItems*4)
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(ResourceLimitError,MemoryAllocationFailed,image);
+              }
               break;
+
             case 0x0E:
               Bitmap2Header1.Width=ReadBlobLSBShort(image);
-              Bitmap2Header1.Heigth=ReadBlobLSBShort(image);
-              if ((Bitmap2Header1.Width == 0) || (Bitmap2Header1.Heigth == 0))
-                ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+              Bitmap2Header1.Height=ReadBlobLSBShort(image);
               Bitmap2Header1.Depth=ReadBlobByte(image);
               Bitmap2Header1.Compression=ReadBlobByte(image);
 
+              if ((Bitmap2Header1.Width == 0) || (Bitmap2Header1.Height == 0))
+              {
+                MagickFreeResourceLimitedMemory(pPalette);
+                ThrowReaderException(CorruptImageError,ImproperImageHeader,image);
+              }
+
+              if(image->rows!=0 && image->columns!=0)
+              {                       /* Allocate next image structure. */
+                if(EnsureNextImage(image_info, &image) < 0)
+                    goto Finish;
+              }
+
               if(Bitmap2Header1.Compression > 1)
                 continue; /*Unknown compression method */
               switch(Bitmap2Header1.Depth)
@@ -1614,7 +1766,28 @@ UnpackRaster:
                   continue;  /*Ignore raster with unknown depth*/
                 }
               image->columns=Bitmap2Header1.Width;
-              image->rows=Bitmap2Header1.Heigth;
+              image->rows=Bitmap2Header1.Height;
+              if (image->logging)
+                (void) LogMagickEvent(CoderEvent,GetMagickModule(),
+                                      "Image dimensions %lux%lu, bpp=%u", image->columns, image->rows, bpp);
+
+              if(pPalette!=NULL && PaletteItems>0)
+              {
+                if(bpp>=16 || PaletteItems<(1<<bpp))
+                  image->colors = PaletteItems;	/*WPG_Palette.NumOfEntries;*/
+                else
+                  image->colors = 1 << bpp;
+                if (!AllocateImageColormap(image,image->colors))
+                  goto NoMemory;
+                image->storage_class = PseudoClass;
+                for (i=0; i<(int)image->colors; i++)
+                {
+                  image->colormap[i].red = ScaleCharToQuantum(pPalette[4*i]);
+                  image->colormap[i].green = ScaleCharToQuantum(pPalette[4*i+1]);
+                  image->colormap[i].blue = ScaleCharToQuantum(pPalette[4*i+2]);
+                  image->colormap[i].opacity = ScaleCharToQuantum(pPalette[4*i+3]);
+                }
+              }
 
               if ((image->colors == 0) && (bpp != 24))
                 {
@@ -1631,7 +1804,6 @@ UnpackRaster:
                         goto NoMemory;
                 }
 
-
               switch(Bitmap2Header1.Compression)
                 {
                 case 0:    /*Uncompressed raster*/
@@ -1708,14 +1880,6 @@ UnpackRaster:
                 if (image->scene >= (image_info->subimage+image_info->subrange-1))
                   goto Finish;
 
-              /* Allocate next image structure. */
-              AllocateNextImage(image_info,image);
-              image->depth=8;
-              if (image->next == (Image *) NULL)
-                goto Finish;
-              image=SyncNextImageInList(image);
-              image->columns=image->rows=0;
-              image->colors=0;
               break;
 
             case 0x12:  /* Postscript WPG2*/
@@ -1736,12 +1900,14 @@ UnpackRaster:
 
     default:
       {
+        MagickFreeResourceLimitedMemory(pPalette);
         ThrowReaderException(CoderError,DataEncodingSchemeIsNotSupported,image);
       }
     }
 
  Finish:
   CloseBlob(image);
+  MagickFreeResourceLimitedMemory(pPalette);
 
   {
     Image
@@ -1753,7 +1919,7 @@ UnpackRaster:
     /*
       Rewind list, removing any empty images while rewinding.
     */
-    p=image;
+    p=image;	//reverted GetFirstImageInList(image);
     image=NULL;
     while (p != (Image *) NULL)
       {
@@ -1779,6 +1945,301 @@ UnpackRaster:
     ThrowReaderException(CorruptImageError,ImageFileDoesNotContainAnyImageData,image);
   return(image);
 }
+
+
+/* RLE helper structure. */
+typedef struct
+{
+	unsigned char count;
+	unsigned char pos;
+	unsigned char buf[254];
+} WPG_RLE_packer;
+
+//FILE *DebugRLE = NULL;
+
+static void WPG_RLE_Flush(WPG_RLE_packer *WPG_RLE, Image *image, unsigned char n)
+{
+  if(n>WPG_RLE->pos) n=WPG_RLE->pos;
+  if(n>0x7F) n=0x7F;
+  if(n>0)
+  {
+    //if(DebugRLE) fprintf(DebugRLE," size=%X",n);
+    WriteBlobByte(image,n);
+    WriteBlob(image, n, WPG_RLE->buf);
+    WPG_RLE->pos -= n;
+    if(WPG_RLE->pos>0)
+      memcpy(WPG_RLE->buf, WPG_RLE->buf+n, n);
+    else
+     WPG_RLE->count = 0;
+  }
+}
+
+
+static void WPG_RLE_AddCharacter(WPG_RLE_packer *WPG_RLE, unsigned char b, Image *image)
+{
+  WPG_RLE->buf[WPG_RLE->pos++] = b;
+
+/*  if(DebugRLE)
+  {
+    fprintf(DebugRLE,"\n%u",b);
+    if(WPG_RLE->pos>=0x7E)
+	fprintf(DebugRLE," *%u %X", WPG_RLE->pos, WPG_RLE->pos);
+  } */
+
+  if(WPG_RLE->pos>1)
+  {
+    if(WPG_RLE->count==0x7E || WPG_RLE->buf[WPG_RLE->pos-2]!=b)
+    {
+      if(WPG_RLE->count>=1)
+      {
+        WPG_RLE->count++;		// True number of repeated BYTEs.
+        WPG_RLE_Flush(WPG_RLE, image, WPG_RLE->pos-1-WPG_RLE->count);
+        WriteBlobByte(image, WPG_RLE->count|0x80);
+        WriteBlobByte(image, WPG_RLE->buf[0]);
+        //if(DebugRLE) fprintf(DebugRLE," count=%X, val=%X",WPG_RLE->count,WPG_RLE->buf[0]);
+        WPG_RLE->pos = 1;
+        WPG_RLE->buf[0] = b;
+      }
+      WPG_RLE->count = 0;
+    }
+    else
+      WPG_RLE->count++;
+  }
+
+  if(WPG_RLE->pos-WPG_RLE->count>0x7E)	// We have uncompressible block with size 0x7F.
+  {
+    WPG_RLE_Flush(WPG_RLE, image, 0x7F);
+    return;
+  }
+  if(WPG_RLE->pos>0x7E && WPG_RLE->count>=1)
+  {
+    WPG_RLE_Flush(WPG_RLE, image, WPG_RLE->pos-1-WPG_RLE->count);
+    return;
+  }
+}
+
+
+static void WPG_RLE_FinalFlush(WPG_RLE_packer *WPG_RLE, Image *image)
+{
+  if(WPG_RLE->count>1)
+  {
+    WPG_RLE_AddCharacter(WPG_RLE, WPG_RLE->buf[WPG_RLE->pos-1]^0xFF, image); // Add a fake BYTE.
+    WPG_RLE->pos = 0;			// Take a last fake BYTE away.
+  }
+  else
+  {
+    WPG_RLE_Flush(WPG_RLE,image,0x7F);
+    WPG_RLE_Flush(WPG_RLE,image,0x7F);
+    WPG_RLE->count = 0;
+  }
+}
+
+
+static void WPG_RLE_AddBlock(WPG_RLE_packer *WPG_RLE, Image *image, const unsigned char *blk, magick_uint16_t size)
+{
+  while(size-- > 0)
+  {
+    WPG_RLE_AddCharacter(WPG_RLE, *blk, image);
+    blk++;
+  }
+}
+
+
+static void WPG_RLE_Init(WPG_RLE_packer *WPG_RLE)
+{
+  WPG_RLE->pos = WPG_RLE->count = 0;
+}
+
+
+/*
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%   W r i t e W P G I m a g e                                           %
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%
+%  Function WriteWPGImage writes an WPG image to a file.
+%
+%  The format of the WriteWPGImage method is:
+%
+%      unsigned int WriteWPGImage(const ImageInfo *image_info,Image *image)
+%
+%  A description of each parameter follows.
+%
+%    o status: Function WriteWPGImage return True if the image is written.
+%      False is returned is there is a memory shortage or if the image file
+%      fails to write.
+%
+%    o image_info: Specifies a pointer to a ImageInfo structure.
+%
+%    o image:  A pointer to an Image structure.
+*/
+static MagickPassFail WriteWPGImage(const ImageInfo *image_info, Image *image)
+{
+  long y;
+  unsigned int status;
+  int logging;
+  unsigned char StoredPlanes;
+  unsigned ldblk;
+  magick_off_t NumericOffs, CurrOffs;
+  QuantizeInfo quantize_info;
+  WPG_RLE_packer PackRLE;
+  unsigned char *pixels;
+
+  /*
+    Open output image file.
+  */
+  assert(image_info != (const ImageInfo *) NULL);
+  assert(image_info->signature == MagickSignature);
+  assert(image != (Image *) NULL);
+  assert(image->signature == MagickSignature);
+  logging = LogMagickEvent(CoderEvent,GetMagickModule(),"enter WPG");
+  status = OpenBlob(image_info,image,WriteBinaryBlobMode,&image->exception);
+  if(status == MagickFail)
+    ThrowWriterException(FileOpenError,UnableToOpenFile,image);
+  WPG_RLE_Init(&PackRLE);
+
+  (void)TransformColorspace(image,RGBColorspace);
+
+  if ((image->storage_class == DirectClass) ||
+     ((image->storage_class == PseudoClass) && (image->colors<=0 || image->colors>256)))
+  {
+    GetQuantizeInfo(&quantize_info);
+    quantize_info.dither = image_info->dither;
+    quantize_info.number_colors = 256;
+    status = QuantizeImage(&quantize_info,image);
+    if(status==MagickFail || image->colors==0) goto ImageFailure;
+  }
+  if(image->colors <= 2)
+  {
+    StoredPlanes=1;
+    ldblk = (image->columns+7)/8;
+  } else if(image->colors <= 16)
+  {
+    StoredPlanes=4;
+    ldblk = (image->columns+1)/2;
+  } else
+  {
+    StoredPlanes = 8;
+    ldblk = image->columns;
+  }
+
+  pixels = MagickAllocateResourceLimitedMemory(unsigned char *,(size_t)(ldblk));
+  if(pixels == (unsigned char *) NULL)
+    ThrowWriterException(ResourceLimitError,MemoryAllocationFailed,image);
+
+	/* Write WPG hader. */
+  WriteBlobLSBLong(image,0x435057FF);		//DWORD FileId
+  WriteBlobLSBLong(image,16);			//DWORD DataOffset;
+  WriteBlobByte(image,1);			//BYTE Product Type
+  WriteBlobByte(image,0x16);			//BYTE FileType;
+  WriteBlobByte(image,1);			//BYTE MajorVersion;
+  WriteBlobByte(image,0);			//BYTE MinorVersion;
+  WriteBlobLSBShort(image,0);			//WORD EncryptKey;
+  WriteBlobLSBShort(image,0);			//WORD Reserved;
+
+	/* Start WPG l1 */
+  WriteBlobByte(image,0xF);
+  WriteBlobByte(image,0x6);
+  WriteBlobByte(image,1);			//BYTE Version number
+  WriteBlobByte(image,0);			//BYTE Flags (bit 0 PostScript, maybe bitmap, bit 1 PostScript, no bitmap
+  WriteBlobLSBShort(image,image->columns);	//WORD Width of image (arbitrary units)
+  WriteBlobLSBShort(image,image->rows);		//WORD Height of image (arbitrary units)
+
+	/* Palette */
+  if(StoredPlanes>1)
+  {
+    magick_uint16_t i;
+    WriteBlobByte(image,0xE);
+    i = 4+3*(1<<StoredPlanes);
+    if(i<0xFF)
+        WriteBlobByte(image,i);
+    else
+    {
+      WriteBlobByte(image,0xFF);
+      WriteBlobLSBShort(image,i);
+    }
+    WriteBlobLSBShort(image,0);				// Start index
+    WriteBlobLSBShort(image,1<<StoredPlanes);		// Num entries
+
+    for(i=0; i<(1<<StoredPlanes); i++)
+    {
+      if(i>=image->colors || image->colormap==NULL)
+      {
+        WriteBlobByte(image,i);
+        WriteBlobByte(image,i);
+        WriteBlobByte(image,i);
+      }
+      else
+      {
+        WriteBlobByte(image,ScaleQuantumToChar(image->colormap[i].red));
+        WriteBlobByte(image,ScaleQuantumToChar(image->colormap[i].green));
+        WriteBlobByte(image,ScaleQuantumToChar(image->colormap[i].blue));
+      }
+    }
+  }
+
+	/* Bitmap 1 header */
+  WriteBlobByte(image,0xB);
+  WriteBlobByte(image,0xFF);
+  NumericOffs = TellBlob(image);
+  WriteBlobLSBShort(image,0x8000);
+  WriteBlobLSBShort(image,0);
+
+  WriteBlobLSBShort(image,image->columns);	//WORD Width
+  WriteBlobLSBShort(image,image->rows);		//WORD Height
+  WriteBlobLSBShort(image,StoredPlanes);	//WORD Depth;
+  WriteBlobLSBShort(image,75);			//WORD HorzRes;
+  WriteBlobLSBShort(image,75);			//WORD VertRes;
+
+  /*
+    Store image data.
+  */
+  for(y=0; y<(long)image->rows; y++)
+  {
+    //if(y==1310 && DebugRLE==NULL) DebugRLE=fopen("o:\\temp\\14\\debug.txt","wb");
+    //if(y>1310 && DebugRLE) {fclose(DebugRLE);DebugRLE=NULL;}
+
+    if(AcquireImagePixels(image,0,y,image->columns,1,&image->exception) == (const PixelPacket *)NULL)
+    {
+      status = MagickFail;
+      break;
+    }
+    if(ExportImagePixelArea(image, (StoredPlanes==1)?GrayQuantum:IndexQuantum, StoredPlanes,pixels,0,0) != MagickPass)
+    {
+      status = MagickFail;
+      break;
+    }
+    WPG_RLE_AddBlock(&PackRLE,image,pixels,ldblk);
+    WPG_RLE_FinalFlush(&PackRLE,image);
+  }
+
+  CurrOffs = TellBlob(image);
+  SeekBlob(image,NumericOffs,SEEK_SET);
+  NumericOffs = CurrOffs - NumericOffs - 4;
+  WriteBlobLSBShort(image, 0x8000|(NumericOffs>>16));
+  WriteBlobLSBShort(image, NumericOffs&0xFFFF);
+  SeekBlob(image,CurrOffs,SEEK_SET);
+
+	/* End of WPG data */
+  WriteBlobByte(image,0x10);
+  WriteBlobByte(image,0);
+
+  MagickFreeResourceLimitedMemory(pixels);
+ImageFailure:
+  status &= CloseBlob(image);
+
+
+  if (logging)
+    (void)LogMagickEvent(CoderEvent,GetMagickModule(),"return WPG");
+
+  return(status);
+}
+
 
 /*
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
@@ -1810,10 +2271,12 @@ ModuleExport void RegisterWPGImage(void)
 
   entry=SetMagickInfo("WPG");
   entry->decoder=(DecoderHandler) ReadWPGImage;
+  entry->encoder = (EncoderHandler)WriteWPGImage;
   entry->magick=(MagickHandler) IsWPG;
   entry->description="Word Perfect Graphics";
   entry->module="WPG";
   entry->seekable_stream=True;
+  entry->adjoin=MagickFalse;
   entry->coder_class=UnstableCoderClass;
   (void) RegisterMagickInfo(entry);
 }
