@@ -1,5 +1,5 @@
 /*
-% Copyright (C) 2003 - 2018 GraphicsMagick Group
+% Copyright (C) 2003-2023 GraphicsMagick Group
 % Copyright (C) 2002 ImageMagick Studio
 % Copyright 1991-1999 E. I. du Pont de Nemours and Company
 %
@@ -37,6 +37,9 @@
   Include declarations.
 */
 #include "magick/studio.h"
+#if defined(USE_GLIBC_MTRACE) && USE_GLIBC_MTRACE
+#  include <mcheck.h>
+#endif /* if defined(USE_GLIBC_MTRACE) && USE_GLIBC_MTRACE */
 #if defined(MSWINDOWS) || defined(__CYGWIN__)
 # include "magick/nt_feature.h"
 #endif
@@ -86,6 +89,8 @@ static SemaphoreInfo
 static MagickInfo
   *magick_list = (MagickInfo *) NULL;
 
+static unsigned int initialize_magick_options = 0;
+
 static unsigned int panic_signal_handler_call_count = 0;
 static unsigned int quit_signal_handler_call_count = 0;
 
@@ -141,10 +146,18 @@ MagickSetFileSystemBlockSize(const size_t block_size)
 %  should be invoked in the primary (original) thread of the application's
 %  process while shutting down, and only after any threads which might be
 %  using GraphicsMagick functions have terminated.  Since GraphicsMagick
-%  uses threads internally via OpenMP, it is also necessary for any function
-%  calls into GraphicsMagick to have already returned so that OpenMP worker
-%  threads are quiesced and won't be accessing any semaphores or data
-%  structures which are destroyed by this function.
+%  may use threads internally via OpenMP, it is also necessary for any
+%  function calls into GraphicsMagick to have already returned so that
+%  OpenMP worker threads are quiesced and won't be accessing any semaphores
+%  or data structures which are destroyed by this function.
+%
+%  The OpenMP implementation (if present) starts/stops any OpenMP worker
+%  threads and allocates/frees OpenMP resources using its own algorithms.
+%  This means that OpenMP worker threads and OpenMP resources are likely
+%  to remain allocated after DestroyMagick() returns.  Since OpenMP 5.0,
+%  invoking omp_pause_resource_all(omp_pause_hard) will assure that any
+%  resources allocated by OpenMP (threads, thread-specific memory, etc.)
+%  are freed, and this may be called after DestroyMagick() has returned.
 %
 %  The format of the DestroyMagick function is:
 %
@@ -176,7 +189,7 @@ DestroyMagick(void)
   DestroyColorInfo();           /* Color database */
   DestroyDelegateInfo();        /* External delegate information */
   DestroyTypeInfo();            /* Font information */
-  DestroyMagicInfo();           /* File format detection */
+  /*DestroyMagicInfo();*/       /* File format detection */
   DestroyMagickInfoList();      /* Coder registrations + modules */
   DestroyConstitute();          /* Constitute semaphore */
   DestroyMagickRegistry();      /* Registered images */
@@ -184,12 +197,15 @@ DestroyMagick(void)
   DestroyMagickRandomGenerator(); /* Random number generator */
   DestroyTemporaryFiles();      /* Temporary files */
 #if defined(MSWINDOWS)
+#if defined(HasGS)
   NTGhostscriptUnLoadDLL();     /* Ghostscript DLL */
+#endif /* if defined(HasGS) */
 #endif /* defined(MSWINDOWS) */
   /*
     Destroy logging last since some components log their destruction.
   */
   DestroyLogInfo();             /* Logging configuration */
+  DestroyMagickExceptionHandling(); /* Exception handling */
   DestroySemaphore();           /* Semaphores framework */
 
   /* Now uninitialized */
@@ -563,20 +579,26 @@ GetMagickInfoArray(ExceptionInfo *exception)
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %
 %  InitializeMagick() initializes the GraphicsMagick environment.
-%  InitializeMagick() MUST be invoked by the using program before making
-%  use of GraphicsMagick functions or else the library will be unusable.
+%
+%  InitializeMagick() or InitializeMagickEx() MUST be invoked by the using
+%  program before making use of GraphicsMagick functions or else the library
+%  will be unusable and any usage is likely to cause a crash.
 %
 %  This function should be invoked in the primary (original) thread of the
 %  application's process, and before starting any OpenMP threads, as part
 %  of program initialization.
 %
+%  If alternate memory allocations are provided via MagickAllocFunctions()
+%  then that function should be invoked before InitializeMagickEx() since
+%  the memory allocation functions need to be consistent.
+%
 %  The format of the InitializeMagick function is:
 %
 %      InitializeMagick(const char *path)
 %
 %  A description of each parameter follows:
 %
-%    o path: The execution path of the current GraphicsMagick client.
+%    o path: The execution path of the current GraphicsMagick client (or NULL)
 %
 %
 */
@@ -602,63 +624,94 @@ static RETSIGTYPE MagickPanicSignalHandler(int signo);
 static RETSIGTYPE MagickSignalHandler(int signo);
 
 /*
-  Signal function which prevents interrupted system calls from
-  automatically being restarted. From W. Richard Stevens "Advanced
-  Programming in the UNIX Environment", Chapter 10.14.
+  Signal function which only sets the signal handler if it its handling
+  is still set to SIG_DFL. This ensures that signal handlers are not
+  registered for signals which are being ignored, or that the API
+  user has already registered before invoking InitializeMagick.
+
+  Due to a weakness in the POSIX APIs, the only way to see the handler
+  which was registered before is by registering a new handler.
+
+  If an API user registers its own signal handler, then it is responsible
+  for invoking DestroyMagick when a signal is received.
 */
-static Sigfunc *
-MagickSignal(int signo, Sigfunc *func)
+static void
+MagickCondSignal(int signo, Sigfunc *func)
 {
 #if defined(HAVE_SIGACTION) && defined(HAVE_SIGEMPTYSET)
   struct sigaction
     act,
     oact;
 
+  /*
+    Signal set-up code which prevents interrupted system calls from
+    automatically being restarted. From W. Richard Stevens "Advanced
+    Programming in the UNIX Environment", Chapter 10.14.
+  */
   act.sa_handler=func;
   (void) sigemptyset(&act.sa_mask);
   act.sa_flags=0;
 #  if defined(SA_INTERRUPT)  /* SunOS */
   act.sa_flags |= SA_INTERRUPT;
-#  endif
+#  endif /* if defined(SA_INTERRUPT) */
 #  if defined(SA_ONSTACK)  /* Use alternate signal stack if available */
   act.sa_flags |= SA_ONSTACK;
-#  endif
-  if (sigaction(signo, &act, &oact) < 0)
-    return (SIG_ERR);
-  return (oact.sa_handler);
+#  endif /* if defined(SA_ONSTACK) */
+
+  (void) memset(&oact, 0, sizeof(oact));
+  if (sigaction(signo, &act, &oact) == 0)
+    {
+      if ((oact.sa_flags & SA_SIGINFO) ||
+          (oact.sa_handler != SIG_DFL))
+        {
+          /* Restore prior handler! */
+          if (sigaction(signo, &oact, &act) != 0)
+            (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                  "Failed to restore prior signal handler for signal ID %d!",signo);
+          else
+            (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                  "Restored prior signal handler for signal ID %d!",signo);
+        }
+      else
+        {
+          (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                "Registered signal handler for signal ID %d",signo);
+        }
+    }
+  else
+    {
+      (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                            "Failed to register signal handler for signal ID %d!",signo);
+    }
+
 #else
-  return signal(signo, func);
-#endif
-}
-
-/*
-  Signal function which only sets the signal handler if it its handling
-  is still set to SIG_DFL. This ensures that signal handlers are not
-  registered for signals which are being ignored, or that the API
-  user has already registered before invoking InitializeMagick.
-
-  If an API user registers its own signal hander, then it is responsible
-  for invoking DestroyMagick when a signal is received.
-*/
-static Sigfunc *
-MagickCondSignal(int signo, Sigfunc *func)
-{
   Sigfunc *
     o_handler;
 
-  if ((o_handler=MagickSignal(signo,func)) != SIG_ERR)
+  if ((o_handler=signal(signo, func)) != SIG_ERR)
     {
-      /*
-        If handler is not the default, then restore the previous
-        setting.
-      */
       if (o_handler != SIG_DFL)
-        (void) MagickSignal(signo,o_handler);
+        {
+          /* Restore prior handler! */
+          if (signal(signo, o_handler) ==  SIG_ERR)
+            (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                  "Failed to restore prior signal handler for signal ID %d!",signo);
+          else
+            (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                  "Restored prior signal handler for signal ID %d!",signo);
+        }
       else
-        (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
-          "Registered signal handler for signal ID %d",signo);
+        {
+          (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                                "Registered signal handler for signal ID %d",signo);
+        }
     }
-  return (o_handler);
+  else
+    {
+      (void) LogMagickEvent(ConfigureEvent,GetMagickModule(),
+                            "Failed to register signal handler for signal ID %d!",signo);
+    }
+#endif
 }
 
 /*
@@ -687,7 +740,7 @@ MagickSignalHandlerMessage(const int signo, const char *subtext)
   */
   static const struct {
     const int signo;
-    const char *name;
+    const char name[8];
     const char *descr;
   } signal_descr[] = {
 #if defined(SIGHUP)
@@ -899,7 +952,7 @@ IsValidFilesystemPath(const char *path)
     {
 #if defined(POSIX)
       /* For POSIX we check the first character to see of it is a file
-          system path seperator. If not then we ignore the passed data
+          system path separator. If not then we ignore the passed data
       */
       if ((*path == *DirectorySeparator))
         return IsAccessibleNoLogging(path);
@@ -925,7 +978,7 @@ IsValidFilesystemPath(const char *path)
 /*
   Try and figure out the path and name of the client
  */
-MagickExport void
+static void
 InitializeMagickClientPathAndName(const char *path)
 {
 #if !defined(UseInstalledMagick)
@@ -1006,60 +1059,122 @@ InitializeMagickClientPathAndName(const char *path)
   Establish signal handlers for common signals
 
 */
-MagickExport void
+static void
 InitializeMagickSignalHandlers(void)
 {
 #if 0
   /* termination of child process */
 #if defined(SIGCHLD)
-  (void) MagickCondSignal(SIGCHLD,MagickIgnoreSignalHandler);
+  MagickCondSignal(SIGCHLD,MagickIgnoreSignalHandler);
 #endif
 #endif
 
   /* hangup, default terminate */
 #if defined(SIGHUP)
-  (void) MagickCondSignal(SIGHUP,MagickSignalHandler);
+  MagickCondSignal(SIGHUP,MagickSignalHandler);
 #endif
   /* interrupt (CONTROL-c), default terminate */
 #if defined(SIGINT)
-  (void) MagickCondSignal(SIGINT,MagickSignalHandler);
+  MagickCondSignal(SIGINT,MagickSignalHandler);
 #endif
   /* quit (CONTROL-\), default terminate with core */
 #if defined(SIGQUIT)
-  (void) MagickCondSignal(SIGQUIT,MagickPanicSignalHandler);
+  MagickCondSignal(SIGQUIT,MagickPanicSignalHandler);
 #endif
   /* software-triggered abort, default terminate with core */
 #if defined(SIGABRT)
-  (void) MagickCondSignal(SIGABRT,MagickPanicSignalHandler);
+  MagickCondSignal(SIGABRT,MagickPanicSignalHandler);
 #endif
   /* floating point exception, default terminate with core */
 #if defined(SIGFPE)
-  (void) MagickCondSignal(SIGFPE,MagickPanicSignalHandler);
+  MagickCondSignal(SIGFPE,MagickPanicSignalHandler);
 #endif
   /* software termination signal from kill, default terminate */
 #if defined(SIGTERM)
-  (void) MagickCondSignal(SIGTERM,MagickSignalHandler);
+  MagickCondSignal(SIGTERM,MagickSignalHandler);
 #endif
   /* Bus Error */
 #if defined(SIGBUS)
-  (void) MagickCondSignal(SIGBUS,MagickPanicSignalHandler);
+  MagickCondSignal(SIGBUS,MagickPanicSignalHandler);
 #endif
   /* segmentation fault */
 #if defined(SIGSEGV)
-  (void) MagickCondSignal(SIGSEGV,MagickPanicSignalHandler);
+  MagickCondSignal(SIGSEGV,MagickPanicSignalHandler);
 #endif
-  /* exceeded cpu limit, default terminate with core */
+  /* exceeded cpu limit, default terminate */
 #if defined(SIGXCPU)
-  (void) MagickCondSignal(SIGXCPU,MagickPanicSignalHandler);
+  MagickCondSignal(SIGXCPU,MagickSignalHandler);
 #endif
-  /* exceeded file size limit, default terminate with core */
+  /* exceeded file size limit, default terminate */
 #if defined(SIGXFSZ)
-  (void) MagickCondSignal(SIGXFSZ,MagickPanicSignalHandler);
+  MagickCondSignal(SIGXFSZ,MagickSignalHandler);
 #endif
 }
 
+
 MagickExport void
 InitializeMagick(const char *path)
+{
+  ExceptionInfo exception;
+  GetExceptionInfo(&exception);
+  (void) InitializeMagickEx(path,0,&exception);
+  if (exception.severity != UndefinedException)
+    CatchException(&exception);
+  DestroyExceptionInfo(&exception);
+}
+
+/*
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%   I n i t i a l i z e M a g i c k E x                                       %
+%                                                                             %
+%                                                                             %
+%                                                                             %
+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
+%
+%  InitializeMagickEx() initializes the GraphicsMagick environment,
+%  providing a bit more more control and visibility over initialization
+%  than the original InitializeMagick().  Use DestroyMagick() to destroy
+%  the GraphicsMagick environment when it is not longer needed.
+%
+%  InitializeMagick() or InitializeMagickEx() MUST be invoked by the using
+%  program before making use of GraphicsMagick functions or else the library
+%  will be unusable and any usage is likely to cause a crash.
+%
+%  This function should be invoked in the primary (original) thread of the
+%  application's process, and before starting any OpenMP threads, as part
+%  of program initialization.
+%
+%  If alternate memory allocations are provided via MagickAllocFunctions()
+%  then that function should be invoked before InitializeMagickEx() since
+%  the memory allocation functions need to be consistent.
+%
+%  Available options are:
+%
+%    o MAGICK_OPT_NO_SIGNAL_HANDER - Don't register ANSI/POSIX signal handlers
+%
+%  The format of the InitializeMagickEx function is:
+%
+%      MagickPassFail InitializeMagickEx(const char *path,
+%                                        unsigned int options,
+%                                        ExceptionInfo *exception)
+%
+%  A description of each parameter follows:
+%
+%    o path: The execution path of the current GraphicsMagick client (or NULL)
+%
+%    o options: Options bit flags tailoring initializations performed
+%
+%    o exception: Information about initialization failure is reported here.
+%
+%
+*/
+
+MagickExport MagickPassFail
+InitializeMagickEx(const char *path, unsigned int options,
+                   ExceptionInfo *exception)
 {
   /*
     NOTE: This routine sets up the path to the client which needs to
@@ -1073,6 +1188,11 @@ InitializeMagick(const char *path)
   const char
     *p;
 
+  MagickPassFail
+    status = MagickPass;
+
+  (void) exception;
+
   /* Acquire initialization lock */
   SPINLOCK_WAIT;
 
@@ -1080,9 +1200,16 @@ InitializeMagick(const char *path)
     {
       /* Release initialization lock */
       SPINLOCK_RELEASE;
-      return;
+      return status;
     }
 
+#if defined(USE_GLIBC_MTRACE) && USE_GLIBC_MTRACE
+  mtrace();
+#endif /* f defined(USE_GLIBC_MTRACE) && USE_GLIBC_MTRACE */
+
+  /* Save initialization options in case we need them later */
+  initialize_magick_options = options;
+
 #if defined(MSWINDOWS)
 # if defined(_DEBUG) && !defined(__BORLANDC__)
   {
@@ -1101,6 +1228,9 @@ InitializeMagick(const char *path)
   /* Initialize semaphores */
   InitializeSemaphore();
 
+  /* Initialize exception handling */
+  InitializeMagickExceptionHandling();
+
   /* Initialize logging */
   InitializeLogInfo();
 
@@ -1169,13 +1299,14 @@ InitializeMagick(const char *path)
 #if defined(MSWINDOWS)
   NTInitializeExceptionHandlers();  /* WIN32 Exceptions */
 #endif /* defined(MSWINDOWS) */
-  InitializeMagickSignalHandlers(); /* Signal handlers */
+  if (!(options & MAGICK_OPT_NO_SIGNAL_HANDER))
+    InitializeMagickSignalHandlers(); /* Signal handlers */
   InitializeTemporaryFiles();       /* Temporary files */
   InitializeMagickResources();      /* Resources */
   InitializeMagickRegistry();       /* Image/blob registry */
   InitializeConstitute();           /* Constitute semaphore */
   InitializeMagickInfoList();       /* Coder registrations + modules */
-  InitializeMagicInfo();            /* File format detection */
+  /*InitializeMagicInfo();*/        /* File format detection */
   InitializeTypeInfo();             /* Font information */
   InitializeDelegateInfo();         /* External delegate information */
   InitializeColorInfo();            /* Color database */
@@ -1192,6 +1323,8 @@ InitializeMagick(const char *path)
 
   /* Release initialization lock */
   SPINLOCK_RELEASE;
+
+  return status;
 }
 
 /*
@@ -1213,7 +1346,7 @@ InitializeMagick(const char *path)
 %
 %
 */
-MagickPassFail
+static MagickPassFail
 InitializeMagickInfoList(void)
 {
   assert(magick_semaphore == (SemaphoreInfo *) NULL);
@@ -1446,23 +1579,24 @@ ListModuleMap(FILE *file,ExceptionInfo *exception)
 MagickExport char *
 MagickToMime(const char *magick)
 {
-  typedef struct _MediaType
+   char
+     media[MaxTextExtent];
+
+  unsigned int
+    i;
+
+  static const struct
   {
     const char
-      *magick,
+      magick[6],
       *media;
-  } MediaType;
-
-  char
-    media[MaxTextExtent];
-
-  MediaType
-    *entry;
-
-  static MediaType
-    MediaTypes[] =
+  }
+  MediaTypes[] =
     {
       { "avi",   "video/avi" },
+      { "apng",  "image/apng" },
+      { "avif",  "image/avif" },
+      { "bmp",   "image/bmp" },
       { "cgm",   "image/cgm;Version=4;ProfileId=WebCGM" }, /* W3 WebCGM */
       { "dcm",   "application/dicom" }, /* Incomplete.  See RFC 3240 */
       { "epdf",  "application/pdf" },
@@ -1477,7 +1611,9 @@ MagickToMime(const char *magick)
       { "g3",    "image/g3fax" },
       { "gif",   "image/gif" },
       { "gif87", "image/gif" },
+      { "ico",   "image/x-icon" },
       { "jpeg",  "image/jpeg" },
+      { "jpg",   "image/jpeg" },
       { "mng",   "video/x-mng" },
       { "mpeg",  "video/mpeg" },
       { "png",   "image/png" },
@@ -1489,12 +1625,12 @@ MagickToMime(const char *magick)
       { "tif",   "image/tiff" },
       { "tiff",  "image/tiff" },
       { "wbmp",  "image/vnd.wap.wbmp" },
-      { (char *) NULL, (char *) NULL }
+      { "webp",  "image/image/webp" }
     };
 
-  for (entry=MediaTypes; entry->magick != (char *) NULL; entry++)
-    if (LocaleCompare(entry->magick,magick) == 0)
-      return(AllocateString(entry->media));
+  for (i=0; i < ArraySize(MediaTypes); i++)
+    if (LocaleCompare(MediaTypes[i].magick,magick) == 0)
+      return(AllocateString(MediaTypes[i].media));
   FormatString(media,"image/x-%.1024s",magick);
   LocaleLower(media+8);
   return(AllocateString(media));
@@ -1683,11 +1819,10 @@ SetMagickInfo(const char *name)
     *magick_info;
 
   assert(name != (const char *) NULL);
-  magick_info=MagickAllocateMemory(MagickInfo *,sizeof(MagickInfo));
+  magick_info=MagickAllocateClearedMemory(MagickInfo *,sizeof(MagickInfo));
   if (magick_info == (MagickInfo *) NULL)
     MagickFatalError3(ResourceLimitFatalError,MemoryAllocationFailed,
       UnableToAllocateMagickInfo);
-  (void) memset(magick_info,0,sizeof(MagickInfo));
   magick_info->name=name;
   magick_info->adjoin=MagickTrue;
   magick_info->raw=MagickFalse;
