@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003 - 2018 GraphicsMagick Group
+ * Copyright (C) 2003 - 2023 GraphicsMagick Group
  * Copyright (C) 2003 ImageMagick Studio
  * Copyright 1991-1999 E. I. du Pont de Nemours and Company
  *
@@ -11,7 +11,7 @@
  * any data corruption problems. This does not verify that the image is
  * correct, only that encode/decode process is repeatable.
  *
- * Written by Bob Friesenhahn <bfriesen@simple.dallas.tx.us>
+ * Written by Bob Friesenhahn <bfriesen@GraphicsMagick.org>
  *
  * The image returned by both reads must be identical in order for the
  * test to pass.
@@ -24,6 +24,29 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <locale.h>
+
+static void DescribeFrames(const ImageInfo *image_info, Image *list, const MagickBool ping)
+{
+  /* [0] AVS 70x46+072 Grayscale 8-bit adea7b1989cc5d19794a25ae3d7d0bc86f83b014f7231a869ee7b97177d54ab5 */
+  static const char descr_fmt[] = "[%s] %m %wx%h%X%y %r %q-bit %#";
+  static const char descr_fmt_ping[] = "[%s] %m %wx%h%X%y %r %q-bit";
+  Image *list_entry = list;
+
+  while (list_entry != (Image *) NULL)
+    {
+      char
+        *text;
+
+      text=TranslateText(image_info,list_entry,ping ? descr_fmt_ping : descr_fmt);
+      if (text != (char *) NULL)
+        {
+          fprintf(stdout,"%s\n", text);
+          MagickFree(text);
+        }
+      list_entry=GetNextImageInList(list_entry);
+    }
+}
 
 int main ( int argc, char **argv )
 {
@@ -63,6 +86,17 @@ int main ( int argc, char **argv )
   ExceptionInfo
     exception;
 
+  /*
+    Initialize locale from environment variables (LANG, LC_CTYPE,
+    LC_NUMERIC, LC_TIME, LC_COLLATE, LC_MONETARY, LC_MESSAGES,
+    LC_ALL), but require that LC_NUMERIC use common conventions.  The
+    LC_NUMERIC variable affects the decimal point character and
+    thousands separator character for the formatted input/output
+    functions and string conversion functions.
+  */
+  (void) setlocale(LC_ALL,"");
+  (void) setlocale(LC_NUMERIC,"C");
+
   if (LocaleNCompare("rwblob",argv[0],7) == 0)
     InitializeMagick((char *) NULL);
   else
@@ -129,6 +163,10 @@ int main ( int argc, char **argv )
             {
               (void) SetLogFormat(argv[++arg]);
             }
+          else if (LocaleCompare("interlace",option+1) == 0)
+            {
+              imageInfo->interlace=StringToInterlaceType(argv[++arg]);
+            }
           else if (LocaleCompare("nocheck",option+1) == 0)
             {
               check=MagickFalse;
@@ -166,19 +204,20 @@ int main ( int argc, char **argv )
   if (arg != argc-2)
     {
       (void) printf("arg=%d, argc=%d\n", arg, argc);
-      (void) printf ( "Usage: %s [-compress algorithm -debug events -depth "
-                      "integer -define value -log format -nocheck -quality quality "
-                      "-size geometry -verbose] "
-                      "infile format\n", argv[0] );
+      (void) printf ( "Usage: %s [-compress algorithm] [-debug events]"
+                      " [-depth integer] [-define value] [-log format]"
+                      " [-interlace interlace] [-nocheck] [-quality quality] "
+                      " [-size geometry] [-verbose]"
+                      " infile format\n", argv[0] );
       (void) fflush(stdout);
       exit_status = 1;
       goto program_exit;
     }
 
-  (void) strncpy(infile, argv[arg], MaxTextExtent);
+  (void) strncpy(infile, argv[arg], MaxTextExtent-1);
   infile[MaxTextExtent-1]='\0';
   arg++;
-  (void) strncpy( format, argv[arg], MaxTextExtent);
+  (void) strncpy( format, argv[arg], MaxTextExtent-1);
   format[MaxTextExtent-1]='\0';
 
   magick_info=GetMagickInfo(format,&exception);
@@ -319,8 +358,8 @@ int main ( int argc, char **argv )
     else
       {
         /* Print a short description of the image to stdout */
-        DescribeImage( ping_image, stdout, MagickFalse );
-        (void) fflush(stdout);
+        /* DescribeFrames(imageInfo, ping_image, MagickTrue); */
+        /* (void) fflush(stdout); */
         DestroyImageList( ping_image );
       }
     if (ping_error)
@@ -416,6 +455,10 @@ int main ( int argc, char **argv )
   MagickFree(blob);
   blob=0;
 
+  /* Print a short description of the image to stdout */
+  DescribeFrames(imageInfo, final, MagickFalse);
+  (void) fflush(stdout);
+
   if (check)
     {
       /*
@@ -441,6 +484,17 @@ int main ( int argc, char **argv )
           (!strcmp( "UYVY", format )) ||
           (!strcmp( "WEBP", format )) ||
           (!strcmp( "YUV", format ))  ||
+
+          (!strcmp( "EPDF", format ))  ||
+          (!strcmp( "EPI", format ))  ||
+          (!strcmp( "EPS", format ))  ||
+          (!strcmp( "EPSF", format ))  ||
+          (!strcmp( "EPSI", format ))  ||
+          (!strcmp( "EPT", format ))  ||
+          (!strcmp( "PDF", format ))  ||
+          (!strcmp( "PS", format ))  ||
+          (!strcmp( "PS2", format ))  ||
+
           (final->compression == JPEGCompression))
         fuzz_factor = 0.06;
 
